#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json

import mock

import trading_backend.exchanges as exchanges


async def create_order_mocked_test_args(
    exchange: exchanges.Exchange,
    exchange_private_post_order_method_name: str,
    exchange_request_referral_key: str,
    should_contains: bool = True,
    result_is_list: bool = False,
    symbol: str = "BTC/USDT",
    amount: int = 1,
    price: int = 1,
    post_order_mock_return_value: dict = None
):
    with mock.patch.object(exchange._exchange.connector.client,
                           "check_required_credentials",
                           mock.Mock(return_value=False)), \
            mock.patch.object(exchange._exchange.connector.client,
                              exchange_private_post_order_method_name,
                              mock.AsyncMock(return_value=post_order_mock_return_value or {})) as post_order_mock:
        # without referral patch
        await exchange._exchange.connector.client.create_limit_buy_order(symbol, amount, price)
        result = post_order_mock.call_args[0][0].get(exchange_request_referral_key, "") \
            if not result_is_list else post_order_mock.call_args[0][0][0].get(exchange_request_referral_key, "")
        if should_contains:
            assert exchange._get_id() not in result
        else:
            assert exchange._get_id() != result

        # with referral patch
        await exchange._exchange.connector.client.create_limit_buy_order(symbol, amount, price,
                                                                         params=exchange.get_orders_parameters())
        result = post_order_mock.call_args[0][0].get(exchange_request_referral_key, "") \
            if not result_is_list else post_order_mock.call_args[0][0][0].get(exchange_request_referral_key, "")
        if should_contains:
            assert exchange._get_id() in result
        else:
            assert exchange._get_id() == result


def get_content(signed_result, in_body):
    if in_body:
        return json.loads(signed_result["body" if in_body else "headers"])
    return signed_result["headers"]


async def sign_test(
    exchange: exchanges.Exchange,
    api,
    broker_id_key: str,
    broker_sign_header_key: str = None,
    section: str = None,
    should_contains: bool = False,
    in_body: bool = False,
    url_path: str = "url/path"
):
    method = "POST"
    params = {}
    headers = None
    body = None
    api_key = "123"
    api_secret = "456"
    password = "789"
    ccxt_client = exchange._exchange.connector.client
    ccxt_client.apiKey = api_key
    ccxt_client.secret = api_secret
    ccxt_client.password = password

    # without referral patch
    if section:
        signed_result = ccxt_client.sign(
            url_path, section=section, method=method, params=params, headers=headers, body=body
        )
    else:
        signed_result = ccxt_client.sign(
            url_path, api=api, method=method, params=params, headers=headers, body=body
        )
    assert signed_result
    content = get_content(signed_result, in_body)
    if should_contains:
        assert exchange._get_id() not in content[broker_id_key]
    else:
        assert content[broker_id_key] != exchange._get_id()
    if broker_sign_header_key:
        assert content[broker_sign_header_key]

    # with referral patch
    exchange.get_orders_parameters()
    if section:
        signed_result = ccxt_client.sign(
            url_path, section=section, method=method, params=params, headers=headers, body=body
        )
    else:
        signed_result = ccxt_client.sign(
            url_path, api=api, method=method, params=params, headers=headers, body=body
        )
    assert signed_result
    content = get_content(signed_result, in_body)
    if should_contains:
        assert exchange._get_id() in content[broker_id_key]
    else:
        assert content[broker_id_key] == exchange._get_id()
    if broker_sign_header_key:
        assert content[broker_sign_header_key]


async def exchange_requests_contains_headers_test(exchange: exchanges.Exchange,
                                                  exchange_header_referral_key: str,
                                                  exchange_header_referral_value: str):
    request_headers = exchange._exchange.connector.client.prepare_request_headers()
    assert request_headers[exchange_header_referral_key] == exchange_header_referral_value



import pytest
import ccxt

import trading_backend.errors


async def check_invalid_account(exchange):
    with pytest.raises(trading_backend.errors.ExchangeAuthError):
        # no API keys
        assert await exchange.is_valid_account()
    with pytest.raises(trading_backend.errors.ExchangeAuthError):
        # ValueError turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY"
        assert await exchange.is_valid_account()


async def check_invalid_account_keys_rights(exchange):
    with pytest.raises(ccxt.AuthenticationError):
        # no API keys
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # ValueError turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY"
        assert await exchange._get_api_key_rights()



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt

import trading_backend.exchanges as exchanges
import trading_backend.enums


class BinanceUS(exchanges.Exchange):
    # todo update this when binanceus broker id is available
    SPOT_ID = None
    MARGIN_ID = None
    FUTURE_ID = None
    REF_ID = None
    IS_SPONSORING = False

    @classmethod
    def get_name(cls):
        return 'binanceus'

    async def _ensure_broker_status(self):
        return f"Broker rebate is not enabled (missing broker id)."

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        # Binance.us specific:
        # It is currently impossible to fetch api key permissions: try to cancel an imaginary order,
        # if a permission error is raised instead of a cancel fail, then trading permissions are missing.
        # updated: 14/04/2024
        try:
            return await self._get_api_key_rights_using_order()
        except ValueError as err:
            raise ccxt.AuthenticationError(f"Invalid key format ({err})")
        # raising 404 error
        # restrictions = await self._exchange.connector.client.sapi_get_account_apirestrictions()
        # rights = []
        # if restrictions.get('enableReading'):
        #     rights.append(trading_backend.enums.APIKeyRights.READING)
        # if restrictions.get('enableSpotAndMarginTrading'):
        #     rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
        #     rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
        #     rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        # if restrictions.get('enableWithdrawals'):
        #     rights.append(trading_backend.enums.APIKeyRights.WITHDRAWALS)
        # return rights

    async def _inner_is_valid_account(self) -> (bool, str):
        return False, await self._ensure_broker_status()



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt.async_support
import pytest

import trading_backend.exchanges as exchanges
import tests.util.create_order_tests as create_order_tests
from tests import gateio_exchange


def test_get_name(gateio_exchange):
    assert exchanges.GateIO(gateio_exchange).get_name() == ccxt.async_support.gateio().id.lower()


@pytest.mark.asyncio
async def test_orders_parameters(gateio_exchange):
    exchange = exchanges.GateIO(gateio_exchange)
    assert exchange.get_headers() == {exchange.HEADER_KEY: exchange._get_id()}
    await create_order_tests.exchange_requests_contains_headers_test(exchange,
                                                                     exchange_header_referral_key=exchange.HEADER_KEY,
                                                                     exchange_header_referral_value=exchange._get_id())



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import ccxt

import trading_backend.exchanges as exchanges
import trading_backend.errors
from tests import default_exchange


def test_get_name(default_exchange):
    assert exchanges.Exchange(default_exchange).get_name() == exchanges.Exchange.get_name()


def test_get_orders_parameters(default_exchange):
    assert exchanges.Exchange(default_exchange).get_orders_parameters({"a": 1}) == {"a": 1}


@pytest.mark.asyncio
async def test_is_valid_account(default_exchange):
    exchange = exchanges.Exchange(default_exchange)
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(return_value=None)) as fetch_balance_mock:
        assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=ccxt.AuthenticationError)) as fetch_balance_mock:
        with pytest.raises(trading_backend.errors.ExchangeAuthError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=KeyError)) as fetch_balance_mock:
        # non ccxt error: proceed to right checks and raise
        with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()
    with mock.patch.object(exchange._exchange.connector.client, "fetch_balance",
                           mock.AsyncMock(side_effect=ccxt.InvalidNonce)) as fetch_balance_mock:
        with pytest.raises(trading_backend.errors.TimeSyncError):
            assert await exchange.is_valid_account() == (True, None)
        fetch_balance_mock.assert_called_once()


@pytest.mark.asyncio
async def test_initialize(default_exchange):
    exchange = exchanges.Exchange(default_exchange)
    init_result = await exchange.initialize()
    assert exchange.get_name().capitalize() in init_result
    assert "Broker" not in init_result


@pytest.mark.asyncio
async def test_ensure_broker_status(default_exchange):
    init_result = await exchanges.Exchange(default_exchange)._ensure_broker_status()
    assert "Broker" in init_result
    assert "enabled" in init_result
